@@ -3,23 +3,18 @@
  *
  *  Interface to Gnome libxslt.
  *
- * Copyright (c) 2001-2002 Zveno Pty Ltd
+ * Copyright (c) 2001-2004 Zveno Pty Ltd
  * http://www.zveno.com/
  *
- * Zveno Pty Ltd makes this software and associated documentation
- * available free of charge for any purpose.  You may make copies
- * of the software but you must include all of this notice on any copy.
+ * See the file "LICENSE" for information on usage and
+ * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * Zveno Pty Ltd does not warrant that this software is error free
- * or fit for any purpose.  Zveno Pty Ltd disclaims any liability for
- * all claims, expenses, losses, damages and costs any user may incur
- * as a result of using, copying or modifying the software.
- *
- * $Id: tclxslt.c,v 1.20 2002/11/29 23:29:24 rnurmi Exp $
+ * $Id: tclxslt.c,v 1.25 2004/07/11 12:17:57 balls Exp $
  *
  */
 
-#include "tclxslt.h"
+#include <tclxslt/tclxslt.h>
+#include <string.h>
 
 #undef TCL_STORAGE_CLASS
 #define TCL_STORAGE_CLASS DLLEXPORT
@@ -28,14 +23,6 @@
 #     include "win/win32config.h"
 #endif
 
-/*
- * For Darwin (MacOS X) in particular, but also others
- */
- 
-#ifndef __WIN32__
-#	define DLLIMPORT EXTERN
-#endif
-
 /*
  * Manage stylesheet objects
  */
@@ -48,8 +35,6 @@ typedef struct TclXSLT_Stylesheet {
   Tcl_Obj *messagecommand;
 } TclXSLT_Stylesheet;
 
-static int ssheetCntr = 0;
-
 /*
  * Extension management
  */
@@ -61,7 +46,12 @@ typedef struct TclXSLT_Extension {
   xsltTransformContextPtr xformCtxt;
 } TclXSLT_Extension;
 
-Tcl_HashTable extensions;
+typedef struct ThreadSpecificData {
+  int initialised;
+  int ssheetCntr;
+  Tcl_HashTable *extensions;
+} ThreadSpecificData;
+static Tcl_ThreadDataKey dataKey;
 
 /*
  * Prototypes for procedures defined later in this file:
@@ -93,6 +83,8 @@ static int TclXSLTExtensionCommand _ANSI_ARGS_((ClientData dummy,
 						int objc,
 						Tcl_Obj *CONST objv[]));
 
+static Tcl_Obj * GetParameters _ANSI_ARGS_((Tcl_Interp *interp,
+					    xsltStylesheetPtr stylesheet));
 static int TclXSLTTransform _ANSI_ARGS_((TclXSLT_Stylesheet *stylesheet,
                                          Tcl_Obj *source,
                                          int paramc,
@@ -148,12 +140,14 @@ typedef struct GenericError_Info {
 static CONST84 char *instanceCommandMethods[] = {
   "cget",
   "configure",
+  "get",
   "transform",
   (char *) NULL
 };
 enum instanceCommandMethods {
   TCLXSLT_CGET,
   TCLXSLT_CONFIGURE,
+  TCLXSLT_GET,
   TCLXSLT_TRANSFORM
 };
 static CONST84 char *instanceCommandOptions[] = {
@@ -166,6 +160,14 @@ enum instanceCommandOptions {
   TCLXSLT_OPTION_METHOD
 };
 
+static CONST84 char *instanceGetMethods[] = {
+  "parameters",
+  (char *) NULL
+};
+enum instanceGetMethods {
+  TCLXSLT_GET_PARAMETERS
+};
+
 static CONST84 char *extensionCommandMethods[] = {
   "add",
   "remove",
@@ -177,110 +179,11 @@ enum extensionCommandMethods {
 };
 
 /*
- * Debugging
+ * libxml2 and libxslt are mostly thread-safe,
+ * but there are issues with error callbacks.
  */
 
-static Tcl_Channel stderrChan;
-static char dbgbuf[200];
-
-/*
-static void DumpTclObj(objPtr)
-    Tcl_Obj *objPtr;
-{
-  Tcl_Obj *elPtr;
-  int idx, len;
-  
-  if (objPtr->typePtr == Tcl_GetObjType("list")) {
-    Tcl_WriteChars(stderrChan, " list(", -1);
-    Tcl_ListObjLength(NULL, objPtr, &len);
-    for (idx = 0; idx < len; idx++) {
-      Tcl_ListObjIndex(NULL, objPtr, idx, &elPtr);
-      DumpTclObj(elPtr);
-    }
-    Tcl_WriteChars(stderrChan, ")", -1);
-  } else if (objPtr->typePtr == Tcl_GetObjType("libxml2-node")) {
-    xmlNodePtr nodePtr = (xmlNodePtr) objPtr->internalRep.otherValuePtr;
-    sprintf(dbgbuf, " nodePtr x%x name \"%s\" value \"%s\"", nodePtr, nodePtr->name, xmlNodeGetContent(nodePtr));
-    Tcl_WriteChars(stderrChan, dbgbuf, -1);
-  } else {
-    sprintf(dbgbuf, " obj x%x \"%s\"", objPtr, Tcl_GetStringFromObj(objPtr, NULL));
-    Tcl_WriteChars(stderrChan, dbgbuf, -1);
-  }
-}
-
-static void DumpTree(nodePtr)
-    xmlNodePtr nodePtr;
-{
-  xmlNodePtr child;
-          switch (nodePtr->type) {
-      case XML_ELEMENT_NODE:
-	sprintf(dbgbuf, "adding element \"%s\" x%x (%s)\n", nodePtr->name, nodePtr, XML_GET_CONTENT(nodePtr));
-        Tcl_WriteChars(stderrChan, dbgbuf, -1);
-        for (child = nodePtr->children; child != NULL; child = child->next) {
-          DumpTree(child);
-        }
-	break;
-      case XML_ATTRIBUTE_NODE:
-	Tcl_WriteChars(stderrChan, "adding attribute\n", -1);
-	break;
-      case XML_TEXT_NODE:
-      case XML_CDATA_SECTION_NODE:
-	sprintf(dbgbuf, "adding textNode \"%s\" x%x\n", XML_GET_CONTENT(nodePtr), nodePtr);
-        Tcl_WriteChars(stderrChan, dbgbuf, -1);
-	break;
-      case XML_ENTITY_REF_NODE:
-	Tcl_WriteChars(stderrChan, "adding entityReference\n", -1);
-	break;
-      case XML_ENTITY_NODE:
-	Tcl_WriteChars(stderrChan, "adding entity\n", -1);
-	break;
-      case XML_PI_NODE:
-	Tcl_WriteChars(stderrChan, "adding processingInstruction\n", -1);
-	break;
-      case XML_COMMENT_NODE:
-	Tcl_WriteChars(stderrChan, "adding comment\n", -1);
-	break;
-      case XML_DOCUMENT_NODE:
-	Tcl_WriteChars(stderrChan, "adding document\n", -1);
-	break;
-      case XML_DOCUMENT_TYPE_NODE:
-	Tcl_WriteChars(stderrChan, "adding docType\n", -1);
-	break;
-      case XML_DOCUMENT_FRAG_NODE:
-	Tcl_WriteChars(stderrChan, "adding documentFragment\n", -1);
-	break;
-      case XML_NOTATION_NODE:
-	Tcl_WriteChars(stderrChan, "adding notation\n", -1);
-	break;
-      case XML_HTML_DOCUMENT_NODE:
-	Tcl_WriteChars(stderrChan, "adding HTMLdocument\n", -1);
-	break;
-      case XML_DTD_NODE:
-	Tcl_WriteChars(stderrChan, "adding dtd\n", -1);
-	break;
-      case XML_ELEMENT_DECL:
-	Tcl_WriteChars(stderrChan, "adding elementDecl\n", -1);
-	break;
-      case XML_ATTRIBUTE_DECL:
-	Tcl_WriteChars(stderrChan, "adding attributeDecl\n", -1);
-	break;
-      case XML_ENTITY_DECL:
-	Tcl_WriteChars(stderrChan, "adding entityDecl\n", -1);
-	break;
-      case XML_NAMESPACE_DECL:
-	Tcl_WriteChars(stderrChan, "adding namespaceDecl\n", -1);
-	break;
-      case XML_XINCLUDE_START:
-	Tcl_WriteChars(stderrChan, "adding xincludeStart\n", -1);
-	break;
-      case XML_XINCLUDE_END:
-	Tcl_WriteChars(stderrChan, "adding xincludeEnd\n", -1);
-	break;
-      default:
-	Tcl_WriteChars(stderrChan, "adding unknown\n", -1);
-        }
-}
-*/
+TCL_DECLARE_MUTEX(libxslt)
 
 /*
  *----------------------------------------------------------------------------
@@ -302,13 +205,18 @@ int
 Xslt_Init (interp)
      Tcl_Interp *interp;	/* Interpreter to initialise */
 {
-  int dbgMode;
+  ThreadSpecificData *tsdPtr;
 
 #ifdef USE_TCL_STUBS
   if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
     return TCL_ERROR;
   }
 #endif
+#ifdef USE_TCLXML_LIBXML2_STUBS
+  if (Tclxml_libxml2_InitStubs(interp, TCLXML_LIBXML2_VERSION, 1) == NULL) {
+    return TCL_ERROR;
+  }
+#endif
 #ifdef USE_TCLDOMXML_STUBS
   /* This is  "dom::libxml2"
    */
@@ -317,14 +225,26 @@ Xslt_Init (interp)
   }
 #endif
 
+  if (Tcl_PkgRequire(interp, "dom", TCLXSLT_VERSION, 0) == NULL) {
+    Tcl_AppendResult(interp, ": unable to load DOM package version \"", TCLXSLT_VERSION, "\"", NULL);
+    return TCL_ERROR;
+  }
+
+  tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
+  tsdPtr->initialised = 1;
+  tsdPtr->ssheetCntr = 0;
+  tsdPtr->extensions = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
+  Tcl_InitHashTable(tsdPtr->extensions, TCL_STRING_KEYS);
+
   Tcl_CreateObjCommand(interp, "xslt::compile", TclXSLTCompileCommand, NULL, NULL);
   Tcl_CreateObjCommand(interp, "xslt::extension", TclXSLTExtensionCommand, NULL, NULL);
 
-  Tcl_InitHashTable(&extensions, TCL_STRING_KEYS);
-
+  Tcl_MutexLock(&libxslt);
   exsltRegisterAll();
+  Tcl_MutexUnlock(&libxslt);
 
-  stderrChan = Tcl_GetChannel(interp, "stderr", &dbgMode);
+  Tcl_SetVar2Ex(interp, "::xslt::libxsltversion", NULL, Tcl_NewStringObj(xsltEngineVersion, -1), 0);
+  Tcl_SetVar2Ex(interp, "::xslt::libexsltversion", NULL, Tcl_NewStringObj(exsltLibraryVersion, -1), 0);
 
   Tcl_PkgProvide(interp, "xslt", TCLXSLT_VERSION);
 
@@ -355,20 +275,25 @@ TclXSLTCompileCommand(dummy, interp, objc, objv)
      int objc;
      Tcl_Obj *CONST objv[];
 {
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   TclXSLT_Stylesheet *info;
   xmlDocPtr origDoc, doc;
   xsltStylesheetPtr ssheetPtr = NULL;
+  void *oldxsltErrorCtx, *oldxmlErrorCtx;
+  xmlGenericErrorFunc old_xsltGenericError, old_xmlGenericError;
   GenericError_Info *errorInfoPtr;
+  Tcl_Obj *errObjPtr = NULL;
 
   if (objc != 2) {
     Tcl_WrongNumArgs(interp, 1, objv, "stylesheet-doc");
     return TCL_ERROR;
   }
 
-  /* Copy the document object, since libxslt clobbers the _private field */
-  if (TclDOM_GetDocFromObj(interp, objv[1], &origDoc) != TCL_OK) {
+  if (TclXML_libxml2_GetDocFromObj(interp, objv[1], &origDoc) != TCL_OK) {
     return TCL_ERROR;
   }
+
+  Tcl_MutexLock(&libxslt);
   doc = xmlCopyDoc(origDoc, 1);
   /*
    * xmlCopyDoc doesn't copy some of the fields.
@@ -382,17 +307,43 @@ TclXSLTCompileCommand(dummy, interp, objc, objv)
    * Prepare for compiling stylesheet
    */
 
+  TclXML_libxml2_ResetError(interp);
+
   errorInfoPtr = (GenericError_Info *) Tcl_Alloc(sizeof(GenericError_Info));
   errorInfoPtr->interp = interp;
   errorInfoPtr->stylesheet = NULL;
   errorInfoPtr->code = TCL_OK;
   errorInfoPtr->msg = NULL;
+  xmlSetGenericErrorFunc((void *) errorInfoPtr,
+			 TclXSLTGenericError);
+
+  /*
+   * Save the previous error context so that it can
+   * be restored upon completion of the operation.
+   */
+  old_xsltGenericError = xsltGenericError;
+  oldxsltErrorCtx = xsltGenericErrorContext;
+  old_xmlGenericError = xmlGenericError;
+  oldxmlErrorCtx = xmlGenericErrorContext;
+
+  xmlSetGenericErrorFunc((void *) errorInfoPtr,
+			 TclXSLTGenericError);
+  xsltSetGenericErrorFunc((void *) errorInfoPtr, TclXSLTGenericError);
 
   /*
    * Compile stylesheet
    */
 
-  if ((ssheetPtr = xsltParseStylesheetDoc(doc)) == NULL) {
+  ssheetPtr = xsltParseStylesheetDoc(doc);
+
+  xmlSetGenericErrorFunc((void *) oldxmlErrorCtx, old_xmlGenericError);
+  xsltSetGenericErrorFunc((void *) oldxsltErrorCtx, old_xsltGenericError);
+
+  Tcl_MutexUnlock(&libxslt);
+
+  errObjPtr = TclXML_libxml2_GetErrorObj(interp);
+
+  if (ssheetPtr == NULL) {
     Tcl_SetResult(interp, "error compiling stylesheet", NULL);
     goto error;
   }
@@ -403,18 +354,15 @@ TclXSLTCompileCommand(dummy, interp, objc, objv)
   }
 
   if (errorInfoPtr->code != TCL_OK) {
-
-    if (errorInfoPtr->msg) {
-      Tcl_SetObjResult(interp, errorInfoPtr->msg);
-    }
-
     goto error;
   }
 
+  /* TODO: notify app of any warnings */
+
   info = (TclXSLT_Stylesheet *) Tcl_Alloc(sizeof(TclXSLT_Stylesheet));
   info->interp = interp;
   info->name = Tcl_Alloc(20);
-  sprintf(info->name, "style%d", ssheetCntr++);
+  sprintf(info->name, "style%d", tsdPtr->ssheetCntr++);
   info->stylesheet = ssheetPtr;
   info->messagecommand = NULL;
 
@@ -426,16 +374,21 @@ TclXSLTCompileCommand(dummy, interp, objc, objv)
 
 error:
 
-  if (errorInfoPtr->msg) {
+  if (errObjPtr) {
+    Tcl_SetObjResult(interp, errObjPtr);
+  } else if (errorInfoPtr->msg) {
+    Tcl_SetObjResult(interp, errorInfoPtr->msg);
     Tcl_DecrRefCount(errorInfoPtr->msg);
   }
   Tcl_Free((char *) errorInfoPtr);
 
+  Tcl_MutexLock(&libxslt);
   if (ssheetPtr) {
     xsltFreeStylesheet(ssheetPtr);
   } else {
     xmlFreeDoc(doc);
   }
+  Tcl_MutexUnlock(&libxslt);
   
   return TCL_ERROR;
 }
@@ -466,9 +419,12 @@ TclXSLTDeleteStylesheet(clientData)
   if (ssheet->messagecommand) {
     Tcl_DecrRefCount(ssheet->messagecommand);
   }
+  Tcl_MutexLock(&libxslt);
   xsltFreeStylesheet(ssheet->stylesheet); /* Also frees document */
+  Tcl_MutexUnlock(&libxslt);
   Tcl_Free((char *) ssheet);
-}
+}
+
 /*
  *----------------------------------------------------------------------------
  *
@@ -573,6 +529,30 @@ TclXSLTInstanceCommand(clientData, interp, objc, objv)
 
     break;
 
+  case TCLXSLT_GET:
+    if (objc != 3) {
+      Tcl_WrongNumArgs(interp, 2, objv, "name");
+      return TCL_ERROR;
+    }
+
+    if (Tcl_GetIndexFromObj(interp, objv[2], instanceGetMethods,
+			    "name", 0, &option) != TCL_OK) {
+      return TCL_ERROR;
+    }
+
+    switch ((enum instanceGetMethods) option) {
+    case TCLXSLT_GET_PARAMETERS:
+
+      Tcl_SetObjResult(interp, GetParameters(interp, ssheet->stylesheet));
+      break;
+
+    default:
+      Tcl_SetResult(interp, "unknown name", NULL);
+      return TCL_ERROR;
+    }
+
+    break;
+
   case TCLXSLT_TRANSFORM:
     if (objc < 3) {
       Tcl_WrongNumArgs(interp, 2, objv, "source ?param value...?");
@@ -618,8 +598,9 @@ TclXSLTTransform(stylesheet, source, paramc, paramv)
   char **params = NULL;
   int nbparams = 0, i;
   GenericError_Info *errorInfoPtr;
-  void *oldErrorCtx;
-  xmlGenericErrorFunc old_xsltGenericError;
+  void *oldxsltErrorCtx, *oldxmlErrorCtx;
+  xmlGenericErrorFunc old_xsltGenericError, old_xmlGenericError;
+  Tcl_Obj *errObjPtr = NULL;
 
   errorInfoPtr = (GenericError_Info *) Tcl_Alloc(sizeof(GenericError_Info));
   errorInfoPtr->interp = stylesheet->interp;
@@ -627,10 +608,12 @@ TclXSLTTransform(stylesheet, source, paramc, paramv)
   errorInfoPtr->code = TCL_OK;
   errorInfoPtr->msg = NULL;
 
-  if (TclDOM_GetDocFromObj(stylesheet->interp, source, &doc) != TCL_OK) {
+  if (TclXML_libxml2_GetDocFromObj(stylesheet->interp, source, &doc) != TCL_OK) {
     goto error;
   }
 
+  TclXML_libxml2_ResetError(stylesheet->interp);
+
   params = (char **) Tcl_Alloc(sizeof(char **) * (paramc + 1));
   for (i = 0; i < paramc; i++) {
     params[nbparams++] = Tcl_GetStringFromObj(paramv[i++], NULL);
@@ -642,6 +625,8 @@ TclXSLTTransform(stylesheet, source, paramc, paramv)
    * Perform the transformation
    */
 
+  Tcl_MutexLock(&libxslt);
+
   /*
    * Save the previous error context so that it can
    * be restored upon completion of the transformation.
@@ -649,27 +634,41 @@ TclXSLTTransform(stylesheet, source, paramc, paramv)
    * recursively (usually due to extensions).
    */
   old_xsltGenericError = xsltGenericError;
-  oldErrorCtx = xsltGenericErrorContext;
+  oldxsltErrorCtx = xsltGenericErrorContext;
+  old_xmlGenericError = xmlGenericError;
+  oldxmlErrorCtx = xmlGenericErrorContext;
 
+  xmlSetGenericErrorFunc((void *) errorInfoPtr,
+			 TclXSLTGenericError);
   xsltSetGenericErrorFunc((void *) errorInfoPtr, TclXSLTGenericError);
 
   result = xsltApplyStylesheet(stylesheet->stylesheet, doc, (const char **)params);
 
-  xsltSetGenericErrorFunc((void *) oldErrorCtx, old_xsltGenericError);
+  xmlSetGenericErrorFunc((void *) oldxmlErrorCtx, old_xmlGenericError);
+  xsltSetGenericErrorFunc((void *) oldxsltErrorCtx, old_xsltGenericError);
+
+  Tcl_MutexUnlock(&libxslt);
+
+  errObjPtr = TclXML_libxml2_GetErrorObj(stylesheet->interp);
 
   if (result == NULL) {
-    Tcl_Obj *resultPtr = Tcl_NewStringObj("no result document", -1);
+    if (errObjPtr) {
+      Tcl_SetObjResult(stylesheet->interp, errObjPtr);
+      goto error;
+    } else {
+      Tcl_Obj *resultPtr = Tcl_NewStringObj("no result document", -1);
 
-    if (errorInfoPtr->msg) {
-      Tcl_AppendToObj(resultPtr, ":\n", -1);
-      Tcl_AppendObjToObj(resultPtr, errorInfoPtr->msg);
+      if (errorInfoPtr->msg) {
+	Tcl_AppendToObj(resultPtr, ":\n", -1);
+	Tcl_AppendObjToObj(resultPtr, errorInfoPtr->msg);
+      }
+
+      Tcl_SetObjResult(stylesheet->interp, resultPtr);
+      goto error;
     }
-
-    Tcl_SetObjResult(stylesheet->interp, resultPtr);
-    goto error;
   }
 
-  if (errorInfoPtr->code != TCL_OK && errorInfoPtr->msg && stylesheet->messagecommand) {
+  if ((errObjPtr || (errorInfoPtr->code != TCL_OK && errorInfoPtr->msg)) && stylesheet->messagecommand) {
 
     /* We have produced a result, but there may possibly
      * have been errors.  Trouble is, there might also
@@ -678,8 +677,12 @@ TclXSLTTransform(stylesheet, source, paramc, paramv)
      */
 
     Tcl_Obj *cmdPtr = Tcl_DuplicateObj(stylesheet->messagecommand);
-    if (Tcl_ListObjAppendElement(stylesheet->interp, cmdPtr, errorInfoPtr->msg) != TCL_OK) {
-      goto error;
+    if (errObjPtr) {
+      Tcl_AppendObjToObj(cmdPtr, errObjPtr);
+    } else {
+      if (Tcl_ListObjAppendElement(stylesheet->interp, cmdPtr, errorInfoPtr->msg) != TCL_OK) {
+	goto error;
+      }
     }
     if (Tcl_GlobalEvalObj(stylesheet->interp, cmdPtr) != TCL_OK) {
       goto error;
@@ -687,13 +690,13 @@ TclXSLTTransform(stylesheet, source, paramc, paramv)
 
   }
 
-  Tcl_SetObjResult(stylesheet->interp, TclDOM_CreateObjFromDoc(result));
+  Tcl_SetObjResult(stylesheet->interp, TclDOM_libxml2_CreateObjFromDoc(stylesheet->interp, result));
 
   if (errorInfoPtr->msg) {
     Tcl_DecrRefCount(errorInfoPtr->msg);
   }
   Tcl_Free((char *) errorInfoPtr);
-  Tcl_Free(params);
+  Tcl_Free((char *) params);
 
   return TCL_OK;
 
@@ -703,13 +706,108 @@ TclXSLTTransform(stylesheet, source, paramc, paramv)
     Tcl_DecrRefCount(errorInfoPtr->msg);
   }
   if (params) {
-    Tcl_Free(params);
+    Tcl_Free((char *) params);
   }
   Tcl_Free((char *) errorInfoPtr);
 
   return TCL_ERROR;
 }
 
+void
+ListObjAppendUniqueList(interp, tablePtr, listPtr, newElementsPtr)
+     Tcl_Interp *interp;
+     Tcl_HashTable *tablePtr;
+     Tcl_Obj *listPtr;
+     Tcl_Obj *newElementsPtr;
+{
+  int len, idx;
+  Tcl_Obj *elementPtr, *keyPtr, *namePtr, *nameURIPtr;
+  Tcl_HashEntry *entryPtr;
+
+  Tcl_ListObjLength(interp, newElementsPtr, &len);
+  for (idx = 0; idx < len; idx++) {
+	Tcl_ListObjIndex(interp, newElementsPtr, idx, &elementPtr);
+	Tcl_ListObjIndex(interp, elementPtr, 0, &namePtr);
+	Tcl_ListObjIndex(interp, elementPtr, 1, &nameURIPtr);
+
+	keyPtr = Tcl_NewObj();
+	Tcl_AppendStringsToObj(keyPtr,
+						   Tcl_GetStringFromObj(nameURIPtr, NULL), 
+						   "^", 
+						   Tcl_GetStringFromObj(namePtr, NULL),
+						   NULL);
+	entryPtr = Tcl_FindHashEntry(tablePtr, (CONST char *) keyPtr);
+	if (entryPtr == NULL) {
+	  Tcl_ListObjAppendElement(interp, listPtr, elementPtr);
+	}
+	Tcl_DecrRefCount(keyPtr);
+  }
+}
+
+/*
+ *----------------------------------------------------------------------------
+ *
+ * GetParameters --
+ *
+ *  Retrieves the parameters for a stylesheet.
+ *
+ * Results:
+ *  Returns a Tcl list object.
+ *
+ * Side effects:
+ *  None.
+ *
+ *----------------------------------------------------------------------------
+ */
+
+static Tcl_Obj *
+GetParameters(interp, stylesheet)
+     Tcl_Interp *interp;
+     xsltStylesheetPtr stylesheet;
+{
+  Tcl_Obj *resultPtr, *objPtr, *keyPtr;
+  xsltStackElemPtr varPtr;
+  Tcl_HashTable entries;  /* to keep track of parameter qnames */
+  int new;
+
+  if (stylesheet == NULL) {
+    return NULL;
+  }
+
+  resultPtr = Tcl_NewListObj(0, NULL);
+  Tcl_InitObjHashTable(&entries);
+
+  for (varPtr = stylesheet->variables; varPtr; varPtr = varPtr->next) {
+    Tcl_Obj *listPtr;
+
+    if (strcmp(varPtr->comp->inst->name, "param") == 0) {
+      listPtr = Tcl_NewListObj(0, NULL);
+      Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(varPtr->name, -1));
+      Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(varPtr->nameURI, -1));
+      Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(varPtr->select, -1));
+
+      Tcl_ListObjAppendElement(interp, resultPtr, listPtr);
+
+	  keyPtr = Tcl_NewStringObj(varPtr->nameURI, -1);
+	  Tcl_AppendStringsToObj(keyPtr, "^", varPtr->name, NULL);
+	  Tcl_CreateHashEntry(&entries, (CONST char *) keyPtr, &new);
+    }
+  }
+
+  objPtr = GetParameters(interp, stylesheet->next);
+  if (objPtr) {
+    ListObjAppendUniqueList(interp, &entries, resultPtr, objPtr);
+  }
+  objPtr = GetParameters(interp, stylesheet->imports);
+  if (objPtr) {
+    ListObjAppendUniqueList(interp, &entries, resultPtr, objPtr);
+  }
+
+  Tcl_DeleteHashTable(&entries);
+
+  return resultPtr;
+}
+
 /*
  *----------------------------------------------------------------------------
  *
@@ -749,11 +847,10 @@ TclXSLTGenericError (void *ctx, const char *msg, ...)
   va_end(args);
 
   if (!errorInfoPtr->interp) {
-    sprintf(dbgbuf, "TclXSLTGenericError: NULL interp, msg \"%s\"\n", buf);
     return;
   }
 
-  if (errorInfoPtr->stylesheet->messagecommand) {
+  if (errorInfoPtr->stylesheet && errorInfoPtr->stylesheet->messagecommand) {
 
     Tcl_Obj *cmdPtr = Tcl_DuplicateObj(errorInfoPtr->stylesheet->messagecommand);
     if (Tcl_ListObjAppendElement(errorInfoPtr->interp, cmdPtr, Tcl_NewStringObj(buf, len)) != TCL_OK) {
@@ -802,6 +899,7 @@ TclXSLTExtensionCommand(dummy, interp, objc, objv)
      int objc;
      Tcl_Obj *CONST objv[];
 {
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   int method, new;
   TclXSLT_Extension *extinfo;
   Tcl_HashEntry *entry;
@@ -824,12 +922,17 @@ TclXSLTExtensionCommand(dummy, interp, objc, objv)
       return TCL_ERROR;
     }
 
+    Tcl_MutexLock(&libxslt);
+
     if (xsltRegisterExtModule(Tcl_GetStringFromObj(objv[2], NULL),
 			      TclXSLTExtInit,
 			      TclXSLTExtShutdown)) {
+      Tcl_MutexUnlock(&libxslt);
       Tcl_SetResult(interp, "cannot register extension module", NULL);
     }
 
+    Tcl_MutexUnlock(&libxslt);
+
     extinfo = (TclXSLT_Extension *) Tcl_Alloc(sizeof(TclXSLT_Extension));
     extinfo->interp = interp;
     extinfo->nsuri = objv[2];
@@ -839,7 +942,7 @@ TclXSLTExtensionCommand(dummy, interp, objc, objv)
 
     extinfo->xformCtxt = NULL;
 
-    entry = Tcl_CreateHashEntry(&extensions, Tcl_GetStringFromObj(objv[2], NULL), &new);
+    entry = Tcl_CreateHashEntry(tsdPtr->extensions, Tcl_GetStringFromObj(objv[2], NULL), &new);
 
     if (!new) {
       Tcl_SetResult(interp, "extension already exists", NULL);
@@ -865,7 +968,7 @@ TclXSLTExtensionCommand(dummy, interp, objc, objv)
      * TODO: Remove previously registered elements and functions.
     */
 
-    entry = Tcl_FindHashEntry(&extensions, Tcl_GetStringFromObj(objv[2], NULL));
+    entry = Tcl_FindHashEntry(tsdPtr->extensions, Tcl_GetStringFromObj(objv[2], NULL));
     if (entry == NULL) {
       Tcl_SetResult(interp, "unknown XML Namespace URI", NULL);
       return TCL_ERROR;
@@ -910,10 +1013,11 @@ TclXSLTExtInit(ctxt, URI)
      xsltTransformContextPtr ctxt;
      const xmlChar *URI;
 {
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   Tcl_HashEntry *entry;
   TclXSLT_Extension *extinfo;
 
-  entry = Tcl_FindHashEntry(&extensions, URI);
+  entry = Tcl_FindHashEntry(tsdPtr->extensions, URI);
   if (entry == NULL) {
     /* Extension module was removed */
     return NULL;
@@ -970,6 +1074,8 @@ TclXSLT_RegisterAll(extinfo, nsuri)
    * reg[1] contains extension functions
    */
 
+  Tcl_MutexLock(&libxslt);
+
   /*
    * First register the extension elements.
    */
@@ -995,6 +1101,7 @@ TclXSLT_RegisterAll(extinfo, nsuri)
 
   ret = Tcl_ListObjLength(extinfo->interp, reg[1], &len);
   if (ret != TCL_OK || len == 0) {
+    Tcl_MutexUnlock(&libxslt);
     return;
   }
 
@@ -1009,6 +1116,8 @@ TclXSLT_RegisterAll(extinfo, nsuri)
     	TclXSLTExtFunction);
   }
 
+  Tcl_MutexUnlock(&libxslt);
+
   Tcl_DecrRefCount(objPtr);
 
   return;
@@ -1062,6 +1171,7 @@ TclXSLTExtElementTransform(ctxt, node, inst, comp)
     xmlNodePtr inst;
     xsltStylePreCompPtr comp; /* unused */
 {
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   TclXSLT_Extension *extinfo;
   Tcl_HashEntry *entry;
   Tcl_Obj *cmdPtr;
@@ -1071,7 +1181,7 @@ TclXSLTExtElementTransform(ctxt, node, inst, comp)
     return;
   }
 
-  entry = Tcl_FindHashEntry(&extensions, inst->ns->href);
+  entry = Tcl_FindHashEntry(tsdPtr->extensions, inst->ns->href);
   if (entry == NULL) {
     /*
      * Cannot find extension module.
@@ -1089,7 +1199,7 @@ TclXSLTExtElementTransform(ctxt, node, inst, comp)
   cmdPtr = Tcl_DuplicateObj(extinfo->tclns);
   Tcl_AppendStringsToObj(cmdPtr, "::", inst->name, NULL);
 
-  if (Tcl_ListObjAppendElement(extinfo->interp, cmdPtr, TclDOM_CreateObjFromNode(node)) != TCL_OK) {
+  if (Tcl_ListObjAppendElement(extinfo->interp, cmdPtr, TclDOM_libxml2_CreateObjFromNode(extinfo->interp, node)) != TCL_OK) {
     Tcl_DecrRefCount(cmdPtr);
     return;
   }
@@ -1139,7 +1249,9 @@ TclXSLTExtFunction(xpathCtxt, nargs)
   TclXSLT_Extension *extinfo;
   Tcl_Obj *cmdPtr, *resultPtr;
   xmlXPathObjectPtr obj;
-  int ret, len;
+  int ret;
+
+  Tcl_MutexLock(&libxslt);
 
   xformCtxt = xsltXPathGetTransformContext(xpathCtxt);
 
@@ -1171,6 +1283,7 @@ TclXSLTExtFunction(xpathCtxt, nargs)
     if (obj == NULL) {
       xmlXPathSetError(xpathCtxt, XPATH_INVALID_OPERAND);
       Tcl_DecrRefCount(cmdPtr);
+      Tcl_MutexUnlock(&libxslt);
       return;
     }
 
@@ -1180,6 +1293,7 @@ TclXSLTExtFunction(xpathCtxt, nargs)
       Tcl_BackgroundError(extinfo->interp);
       Tcl_DecrRefCount(objv[0]);
       Tcl_DecrRefCount(cmdPtr);
+      Tcl_MutexUnlock(&libxslt);
       return;
     }
 
@@ -1214,6 +1328,8 @@ TclXSLTExtFunction(xpathCtxt, nargs)
     xpathCtxt->error = XPATH_UNKNOWN_FUNC_ERROR;
   }
 
+  Tcl_MutexUnlock(&libxslt);
+
   Tcl_DecrRefCount(resultPtr);
 
 }
@@ -1227,6 +1343,8 @@ TclXSLTExtFunction(xpathCtxt, nargs)
  *  Data type is preserved, with nodesets being
  *  mapped from a list of nodes.
  *
+ * NB. Mutex is assumed to be locked when invoking this routine.
+ *
  * Results:
  *  XPath Object.
  *
@@ -1244,14 +1362,11 @@ TclXSLT_ConvertTclObjToXPathObj(interp, objPtr)
   xmlNodePtr nodePtr;
   xmlDocPtr docPtr;
 
-  if (TclDOM_GetNodeFromObj(interp, objPtr, &nodePtr) == TCL_OK) {
+  if (TclDOM_libxml2_GetNodeFromObj(interp, objPtr, &nodePtr) == TCL_OK) {
     return xmlXPathNewNodeSet(nodePtr);
   }
   
-  /*
-   * BUG: This is corrupting objPtr for some unknown reason.
-   */
-   if (TclDOM_GetDocFromObj(interp, objPtr, &docPtr) == TCL_OK) {
+   if (TclXML_libxml2_GetDocFromObj(interp, objPtr, &docPtr) == TCL_OK) {
     return xmlXPathNewNodeSet((xmlNodePtr) docPtr);
 
   }
@@ -1292,10 +1407,10 @@ TclXSLT_ConvertTclObjToXPathObj(interp, objPtr)
      * First pass: check that the elements are all nodes.
      */
     for (i = 0; i < len; i++) {
-      if (TclDOM_GetDocFromObj(interp, listPtr[i], &docPtr) == TCL_OK) {
+      if (TclXML_libxml2_GetDocFromObj(interp, listPtr[i], &docPtr) == TCL_OK) {
         continue;
       }
-      if (TclDOM_GetNodeFromObj(interp, listPtr[i], &nodePtr) != TCL_OK) {
+      if (TclDOM_libxml2_GetNodeFromObj(interp, listPtr[i], &nodePtr) != TCL_OK) {
         return xmlXPathNewString(Tcl_GetStringFromObj(objPtr, NULL));
       }
     }
@@ -1303,17 +1418,17 @@ TclXSLT_ConvertTclObjToXPathObj(interp, objPtr)
      * Now go ahead and create the nodeset (we already did the hard
      * work to create internal reps in pass 1).
      */
-    if (TclDOM_GetDocFromObj(interp, listPtr[0], &docPtr) == TCL_OK) {
+    if (TclXML_libxml2_GetDocFromObj(interp, listPtr[0], &docPtr) == TCL_OK) {
       nset = xmlXPathNodeSetCreate((xmlNodePtr) docPtr);
     } else {
-      TclDOM_GetNodeFromObj(interp, listPtr[0], &nodePtr);
+      TclDOM_libxml2_GetNodeFromObj(interp, listPtr[0], &nodePtr);
       nset = xmlXPathNodeSetCreate(nodePtr);
     }
     for (i = 1; i < len; i++) {
-      if (TclDOM_GetDocFromObj(interp, listPtr[i], &docPtr) == TCL_OK) {
+      if (TclXML_libxml2_GetDocFromObj(interp, listPtr[i], &docPtr) == TCL_OK) {
         xmlXPathNodeSetAdd(nset, (xmlNodePtr) docPtr);
       } else {
-        TclDOM_GetNodeFromObj(interp, listPtr[i], &nodePtr);
+        TclDOM_libxml2_GetNodeFromObj(interp, listPtr[i], &nodePtr);
         xmlXPathNodeSetAdd(nset, nodePtr);
       }
     }
@@ -1355,10 +1470,17 @@ TclXSLT_ConvertXPathObjToTclObj(interp, xpobj)
     case XPATH_NODESET:
 
       objPtr = Tcl_NewListObj(0, NULL);
-      for (i = 0; i < xpobj->nodesetval->nodeNr; i++) {
-        Tcl_Obj *nodeObjPtr;
-        nodeObjPtr = TclDOM_CreateObjFromNode(xpobj->nodesetval->nodeTab[i]);
-        Tcl_ListObjAppendElement(interp, objPtr, nodeObjPtr);
+      if (xpobj->nodesetval) {
+	for (i = 0; i < xpobj->nodesetval->nodeNr; i++) {
+	  Tcl_Obj *nodeObjPtr = NULL;
+	  if (xpobj->nodesetval->nodeTab[i] &&
+	      xpobj->nodesetval->nodeTab[i]->type == XML_DOCUMENT_NODE) {
+	    nodeObjPtr = TclXML_libxml2_CreateObjFromDoc((xmlDocPtr) xpobj->nodesetval->nodeTab[i]);
+	  } else if (xpobj->nodesetval->nodeTab[i]) {
+	    nodeObjPtr = TclDOM_libxml2_CreateObjFromNode(interp, xpobj->nodesetval->nodeTab[i]);
+	  }
+	  Tcl_ListObjAppendElement(interp, objPtr, nodeObjPtr);
+	}
       }
 
       break;
